// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include the GUI:
#define IMGUI_DEFINE_MATH_OPERATORS
#include <imgui_internal.h>
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

// Include functions to display Mat4f and Vec4f in the GUI:
# include <imgui_cg1_helpers.h>

// Include OpenGL3.3 Core functions:
#include <glad/glad.h>

// Include GLFW/OpenGL:
#include <GLFW/glfw3.h>

// Include IO utilities:
#include <iostream>

// Include Vec4f class:
#include "src/math/Vec4f.h"

// Include Mat4f class:
#include "src/math/Mat4f.h"

// Include our shader wrapper class:
#include "src/Shader.h"

// Include our mesh class:
#include "src/Mesh.h"

/**
 * Initializes the application including the GUI
 */
int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(860, 560, "Computergrafik 1", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;     // Enable Keyboard Controls

    // Setup Dear ImGui style:
    ImGui::StyleColorsDark();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Defines the with of the left menu:
    int menuWidth = 260;

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf", 16.0f);

    // Enable depth test:
    glEnable(GL_DEPTH_TEST);

    // Ghost Teapot data:
    bool showGhostTeapots = true;
    Vec4f ghostColor(0.6f, 0.7f, 0.8f, 1.0f);
    Mat4f ghostMatrix1 = Mat4f({0.24997943f,-2.185e-08f,1.185e-09f,-8.185e-09f,-2.185e-08f,0.7499864f,-2.912e-08f,-6.658e-09f,-2.912e-08f,-6.658e-09f,0.24997943f,1.185e-09f,-0.75f,0.6f,0.0000002f,1.f});
    Mat4f ghostMatrix2 = Mat4f({-2.185e-08f,0.f,-0.4999f,0.f,-0.4999f,-2.185e-08f,2.185e-08f,0.f,-2.185e-08f,0.4999f,9.552e-16f,0.f,0.8f,-0.6f,0.f,1.f});

    // Create our mesh containing one triangle:
    Mesh triangleMesh(CMAKE_SOURCE_DIR "/models/teapot_scaled.obj");

    // Load our teapot shader:
    Shader shader(CMAKE_SOURCE_DIR "/shader/teapot.vert", CMAKE_SOURCE_DIR "/shader/teapot.frag");

    // These variables sum the mouse movements in x and y direction
    // while the mouse was pressed in window coordinates (i.e. pixels):
    float mouseMovementX = 0.f;
    float mouseMovementY = 0.f;

    // Last mouse position:
    ImVec2 lastMousePosition = ImGui::GetMousePos();

    // Should use perspective projection?:
    bool usePerspectiveTransformation = false;

    // Color of teapot:
    Vec4f colorOfTeapot(0.784f, 0.549f, 0.275f);

    // Position of teapot:
    Vec4f positionOfTeapot(0.0, 0.0, 0.0);

    // Transform Mode for exercise 2:
    int transformMode = 0;

    // Main loop which is executed every frame until the window is closed:
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Subtract right menu:
        display_w -= menuWidth;

        // Create the GUI:
        {
            // Start the Dear ImGui frame
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
            ImGui::NewFrame();

            // Create Window with settings:
            ImGui::SetNextWindowPos(ImVec2(0,0));
            ImGui::SetNextWindowSize(ImVec2(float(menuWidth),float(display_h)));
            ImGui::Begin("Einstellungen", nullptr, ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoCollapse | ImGuiWindowFlags_NoResize);
            ImGui::Text("colorOfTeapot = ");
            ImGui::ColorEdit3("Choose color for colorOfTeapot-Variable:", &colorOfTeapot.x, ImGuiColorEditFlags_NoLabel | ImGuiColorEditFlags_Float);

            ImGui::Separator();
            ImGui::Text("");
            ImGui::Separator();
            ImGui::Text("Aufgabe 2");
            ImGui::Separator();
            ImGui::Checkbox("Show Ghost Teapots", &showGhostTeapots);
            ImGui::Separator();
            ImGui::RadioButton("Default", &transformMode, 0);
            ImGui::RadioButton("Teapot 1 (a)", &transformMode, 1);
            ImGui::RadioButton("Teapot 1 (b)", &transformMode, 2);
            ImGui::RadioButton("Teapot 2 (a)", &transformMode, 3);
            ImGui::RadioButton("Teapot 2 (b)", &transformMode, 4);

            ImGui::Separator();
            ImGui::Text("");
            ImGui::Separator();
            ImGui::Text("Aufgabe 3");
            ImGui::Separator();
            ImGui::Text("positionOfTeapot = ");
            ImGuiCG1::DragVec4f("", &positionOfTeapot.x, true);
            ImGui::Separator();

            ImGui::Text("");
            ImGui::Separator();
            ImGui::Text("Aufgabe 4");
            ImGui::Separator();
            ImGui::Checkbox("Perspektivische Transformation", &usePerspectiveTransformation);

            ImGui::End();
        }

        // Setup the GL viewport
        {
            int fbWidth, fbHeight;
            int winWidth, winHeight;
            glfwGetFramebufferSize(window, &fbWidth, &fbHeight);
            glfwGetWindowSize(window, &winWidth, &winHeight);
            float pixelRatio = (float)fbWidth / (float)winWidth;
            glViewport(menuWidth * pixelRatio, 0, display_w - menuWidth * (pixelRatio-1), display_h);
        }

        glClearColor(0.6f, 0.725f, 0.8f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Calculate aspect ratio (might be used for perspective transformation):
        float aspectRatio = float(display_w) / display_h;

        // Ensure that both sides of the triangle are rendered:
        glDisable(GL_CULL_FACE);

        // If mouse is down, simply add the difference to mouseMovementX and mouseMovementY:
        if(ImGui::IsMouseDown(ImGuiMouseButton_Left) && !ImGui::GetIO().WantCaptureMouse){
            // Calculate delta between current mouse postion and the mouse position in the
            // last frame:
            ImVec2 mouseDelta = ImGui::GetMousePos() - lastMousePosition;

            // Add the delta to mouseMovementX and mouseMovementY:
            mouseMovementX += mouseDelta.x;
            mouseMovementY += mouseDelta.y;
        }

        // *TODO* (Exercise 3): (a) Create a rotation matrix by using Mat4f::rotationX and Mat4f::rotationY,
        //                      (b) add an translation
        Mat4f model;

        // *TODO* (Exercise 2, Teapot 1 (a)):
        if(transformMode == 1){
            model = Mat4f();
        }

        // *TODO* (Exercise 2, Teapot 1 (b)):
        if(transformMode == 2){
            model = Mat4f();
        }

        // *TODO* (Exercise 2, Teapot 2 (a)):
        if(transformMode == 3){
            model = Mat4f();
        }

        // *TODO* (Exercise 2, Teapot 2 (b)):
        if(transformMode == 4){
            model = Mat4f();
        }

        // Initially, we simply have an identity matrix for the projection, so that it
        // does not affect the position when it is multiplied with other matrices:
        Mat4f projection;

        if(usePerspectiveTransformation){
            // *TODO* (Exercise 4): Use projection matrix and describe the visible results.
        }

        // Store this mouse position for the next frame:
        lastMousePosition = ImGui::GetMousePos();

        // Bind our shader:
        shader.bind();

        // Upload the model matrix (used in vertex shader):
        shader.setUniform("model", model);

        // Upload the projection matrix (used in vertex shader):
        shader.setUniform("projection", projection);

        // Upload our colorOfTeapot-Variable (used in fragment shader):
        shader.setUniform("colorOfTeapot", colorOfTeapot);

        // Draw our triangle:
        triangleMesh.render();

        if(showGhostTeapots){
            shader.setUniform("colorOfTeapot", ghostColor);

            shader.setUniform("model", ghostMatrix1);
            triangleMesh.render();

            shader.setUniform("model", ghostMatrix2);
            triangleMesh.render();
        }

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}

